#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2008, 2017
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  dts.tcl
#
# This file will generate a device tree source file that represents
# the Openfirmware devtree tha Mambo created
#
# This can be used by the Device Tree Compiler:
#   http://www.jdl.com/software/)
# which can be used to generate a flattened device tree use by simulated
# software
#

namespace eval dts {

proc get_driver_name { path } {
    set i [ string last "/" $path ]
    if { 0 == $i  && [ string length $path ] == 1 }  then {
        set name "/"
    } else {
        incr i
        set name [ string range $path $i end ]
    }
    return $name;
}

# We "fixup" properties, some we drop (by returning false) but the
# most difficult is we need to identify properties with <phandle>s and
# fix the up. Unfortunately, there is no automagic way to figure these
# out.

proc fix_prop { dev_name prop prop_type prop_value } {
    switch -regexp $prop {
	(^name$) { return false }
    }
    switch -regexp $dev_name {
	(^memory.*$) {
	    switch -regexp $prop {
		(^available$) { return false }
	    }
	}
	(^chosen$) {
	    switch -regexp $prop {
		(^bootargs$) { return true }
		default { return false }
	    }
	}
	(^cpu@[0-9a-f]+$) {
	    switch -regexp $prop {
		(^next-level-cache$) {
		    uplevel {
			set prop_value [format "&node_0x%x" $prop_value]
			set prop_type "phandle"
		    }
		    return true
		}
	    }
	}
    }
    return true
}

proc show_prop { dev_name prop prop_value prop_type pad chid } {

    if { ! [fix_prop $dev_name $prop $prop_type $prop_value] } { return }

    switch -regexp $prop_type {
	(^phandle$) {
	    puts $chid [format "%s%s = < %s >;" $pad $prop $prop_value]
	}
	(^int$) {
	    puts $chid [format "%s%s = < %s >;" $pad $prop $prop_value]
	}
	(^str$) {
	    puts $chid [format "%s%s = \"%s\";" $pad $prop $prop_value]
	}
	(^byte_array$) {
	    puts -nonewline $chid [format "%s%s = \[ " $pad $prop]

	    for { set i 0 } { $i < [ llength $prop_value ] } { incr i } {
		puts -nonewline $chid [format "%s " [ lindex $prop_value $i ] ]
	    }
	    puts $chid "];"
	}
	(^enum$|^array$) {
	    puts -nonewline $chid [format "%s%s = \< " $pad $prop]

	    for { set i 0 } { $i < [ llength $prop_value ] } { incr i } {
		puts -nonewline $chid [format "%s " [ lindex $prop_value $i ] ]
	    }
	    puts $chid ">;"
	}
	(^array64$) {
	    puts -nonewline $chid [format "%s%s = \< " $pad $prop]

	    for { set i 0 } { $i < [ llength $prop_value ] } { incr i } {
		set val [format "%s" [lindex $prop_value $i]]
		set val1 [expr $val >> 32]
		set val2 [expr $val & 0xffffffff]
		puts -nonewline $chid "0x$val1 0x$val2 "
	    }
	    puts $chid ">;"
	}
	(^empty$|^none$) {
	    puts $chid [format "%s%s;" $pad $prop]
	}
    }
}

proc show_props { sim dev_name dev_node indent chid } {
    set prop [ $sim mcm 0 of nextprop $dev_node ]
    set count 0
    set pad [string repeat " " [ expr $indent + 4 ] ]
    while { $prop != "@null" } {
        set prop_value [ $sim mcm 0 of getprop $dev_node $prop ]
        set prop_type [ $sim mcm 0 of getproptype $dev_node $prop ]
        show_prop $dev_name $prop $prop_value $prop_type $pad $chid
        set prop [ $sim mcm 0 of nextprop $dev_node $prop ]
        incr count
        if { $count > 500 } {
	    puts "Endless array: $devname"
	    exit 1
        }
    }
}

proc show_device { sim dev_node indent chid } {
    set name [ $sim mcm 0 of package_to_path $dev_node ]
    set dev_name [get_driver_name $name]
    puts $chid ""
    if { $dev_name != "/" } {
	set node_str [format "node_0x%x: %s" 0x$dev_node $dev_name]
    } else {
	set node_str $dev_name
    }
    puts $chid "[string repeat " " $indent]$node_str \{"
    show_props $sim $dev_name $dev_node $indent $chid
    set child [$sim mcm 0 of child $dev_node ]
    while { $child != 0 } {
        show_device $sim $child [ expr $indent + 4 ] $chid
        set child [ $sim mcm 0 of peer $child ]
    }
    puts $chid "[string repeat " " $indent]\};"
}


proc show_tree { { sim mysim }
		 { chid stdout }
		 { rlist { } } 
	     } {
    set dev "/"
    set dev_node [ $sim mcm 0 of find_device $dev ]
    if { $dev_node <= 0 } {
        puts "device $dev not found"
        return
    }

    puts $chid "/dts-v1/;"
    puts $chid ""
    for { set i 0 } { $i < [ llength $rlist ] } { incr i } {
	set start [lindex $rlist $i 0 ]
	set size [lindex $rlist $i 1 ]
	puts $chid [format "/memreserve/ 0x%x 0x%s " $start $size]
    }

    if { [catch {show_device $sim $dev_node 0 $chid} errs] } {
        puts "show_device got error $errs\n"
    }
}

proc write_dts { sim } {
    # generate temp file
    if { [info exists env(TMPDIR)] } { 
	set tmpdir $env(TMPDIR)
    } {
	set tmpdir /tmp
    }

    set tmp [file join $tmpdir dts.[pid]]
    set dts $tmp.dts
    set dtb $tmp.dtb

    set out [open $dts w+]
    show_tree $sim $out
    close $out

    return $dts
}

proc generate_dtb { dtc dts } {
    if { ![file executable $dtc] } {
	# find the DTC executable
	# in ENV
	set path_list $::env(PATH)
	if { [info exists env(DTC)] } {
	    set dtc ::$env(DTC)
	} {
	    set path [split $path_list :]
	    foreach p $path {
		set e [file join  $p dtc]
		if { [file executable $e] } {
		    set dtc $e
		    break
		}
	    }
	}
    }

    if { ![file executable $dtc] } {
	puts "Can not find the dtc executable"
	exit 1
    }
	    
    puts "Using device tree compiler: $dtc"

    # generate temp file
    if { [info exists env(TMPDIR)] } { 
	set tmpdir $env(TMPDIR)
    } {
	set tmpdir /tmp
    }

    set tmp [file join $tmpdir dts.[pid]]
    set dtb $tmp.dtb

    if { [catch { exec $dtc -q -I dts -O dtb -o $dtb $dts 2>&1 }] } {
	puts "FAIL: $dtc -q -I dts -O dtb -o $dtb $dts"
	return false
    }

    return $dtb
}


proc load_dtb { sim dtb mem } {
    set dtb_size [file size $dtb]

    set ret [$sim memory fread $mem $dtb_size $dtb]

    $sim cpu 0 set gpr 3 $mem
    $sim cpu 0 set gpr 4 0
    $sim cpu 0 set gpr 5 0
    # ePAPR booke magic number
    $sim cpu 0 set gpr 6 0x45504150
    $sim cpu 0 set gpr 7 0
    $sim cpu 0 set gpr 8 0
    $sim cpu 0 set gpr 9 0

    return true
}

proc epapr { sim dtc mem } {

    set dts [write_dts $sim]
    set dtb [generate_dtb $dtc $dts]
    load_dtb $sim $dtb $mem
    return true
}
}
