#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2003, 2017
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#       IBM Corp.
#
#------------------------------------------------------------------------
#  qtrace.tcl
#      This package contains routines for managing the qtrace plugin
#------------------------------------------------------------------------

#------------------------------------------------------------------------
# QTrace namespace
#     This namespace holds information and functions related to qtrace
#------------------------------------------------------------------------
namespace eval QTrace {
    variable version    "1.0"
    variable plugin     ""
    variable loaded     0
    variable generating 0
    variable filename   ""
    variable dispatch_filter
    variable dispatch_focus
    variable dispatch_callback_id
    variable priv_filter
    variable priv_mode
    variable priv_trig_id
    variable debug      0
    variable override
}

package provide QTrace $QTrace::version
package require BCast

#------------------------------------------------------------------------
# QTrace::Initialize
#------------------------------------------------------------------------
proc QTrace::Initialize { sys {sim mysim} } {
    global env
    if {[llength [array names env SYSTEMSIM_TOP]] == 0} {
         if {[catch {open "|which systemsim | cat"} rc]} {
             puts "QTrace::Initialize 'which systemsim' failed to return anything"
             exit
         } else {
             if {[gets $rc path] <= 0} {
                 # systemsim not found in path, guess 3-up from here...
                 set env(SYSTEMSIM_TOP) "[pwd]/../../.."
                 puts "QTrace::Initialize: WARNING -- unable to find systemsim and SYSTEMSIM_TOP not set..."
                 puts "QTrace::Initialize: WARNING -- guessing SYSTEMSIM_TOP should be set to $env(SYSTEMSIM_TOP)"
             } else {
                 # set SYSTEMSIM_TOP to directory where systemsim was found
                 set env(SYSTEMSIM_TOP) "[file dirname $path]/.."
             }
        }
    }

    if {[llength [array names env LIB_DIR]] == 0} {
        set env(LIB_DIR) $env(SYSTEMSIM_TOP)/lib
    }

    variable generating 
    if {$generating} {
        variable filename
        puts "QTrace::Initialize called while generating to file '$filename'..."
        puts "QTrace::Initialize stopping trace capture currently in effect"
        QTrace::Finish $sim
    }

    variable loaded
    if {$loaded} {
        variable plugin
        puts "QTrace::Initialize called while plugin '$plugin' loaded..."
        puts "QTrace::Unloading plugin '$plugin'"
        QTrace::Finish $sim
    }

    variable plugin $env(LIB_DIR)/libqtbin-${sys}.so
}

#------------------------------------------------------------------------
# QTrace::Warm
#------------------------------------------------------------------------
proc QTrace::Warm { fn {sim mysim} } {
    variable plugin 
    variable loaded

    if {! $loaded} {
        $sim plugin load normal $plugin
        set loaded 1
    }
    $sim plugin signal $plugin warm ${fn}.wl
}

#------------------------------------------------------------------------
# QTrace::Start
#------------------------------------------------------------------------
proc QTrace::Start { fn {sim mysim} {mode normal} } {
    variable generating

    if {! $generating} {
        variable plugin 
        variable loaded 

        if {! $loaded} {
            $sim plugin load normal $plugin
            set loaded 1
        }

	    switch -exact $mode {
            "debug" {
                    $sim plugin signal $plugin start ${fn}.qt. txt ${fn}.txt. pte ${fn}.pte.
            }
            default {
                    $sim plugin signal $plugin start ${fn}.qt. pte ${fn}.pte.
            }
        }

        set filename $fn
        set generating 1
    }
}

#------------------------------------------------------------------------
# QTrace::Stop
#------------------------------------------------------------------------
proc QTrace::Stop { {sim mysim} } {
    variable generating

    if {$generating} {
        variable plugin 
        $sim plugin signal $plugin stop
        set generating 0
    }
}

#------------------------------------------------------------------------
# QTrace::Pause
#------------------------------------------------------------------------
proc QTrace::Pause { {thread all} {cpu all} {sim mysim} } {
    variable generating

    if {$generating} {
        variable plugin 
		if {[string eq $cpu "all"]} {
			$sim plugin signal $plugin pause
		} elseif {[string eq $thread "all"]} {
			$sim plugin signal $plugin pause cpu $cpu
		} else {
			$sim plugin signal $plugin pause cpu $cpu thread $thread
		}

	}
    #puts "QTrace::Pause thread $thread cpu $cpu sim $sim"
    #simdebug set instruction 0
}

#------------------------------------------------------------------------
# QTrace::Resume
#------------------------------------------------------------------------
proc QTrace::Resume { {thread all} {cpu all} {sim mysim} } {
    variable generating

    if {$generating} {
        variable plugin 
		if {[string eq $cpu "all"]} {
			$sim plugin signal $plugin resume
		} elseif {[string eq $thread "all"]} {
			$sim plugin signal $plugin resume cpu $cpu
		} else {
			$sim plugin signal $plugin resume cpu $cpu thread $thread
		}
    }
    #puts "QTrace::Resume thread $thread cpu $cpu sim $sim"
    #simdebug set instruction 1
}

#------------------------------------------------------------------------
# QTrace::Finish
#------------------------------------------------------------------------
proc QTrace::Finish { {sim mysim} } {
    variable plugin 
    variable generating
    variable loaded

    if {$generating} {
        $sim plugin signal $plugin stop
        set generating 0
    }
    catch {$sim plugin unload $plugin}
    set loaded 0
}

#------------------------------------------------------------------------
# QTrace::Filter
#------------------------------------------------------------------------
proc QTrace::Filter {mct} {
    variable dispatch_filter
    variable priv_filter
    variable override
    variable debug

    if {![regexp {((\d):)?(\d):(\d)} $mct match x m c t]} {
        return -code error "QTrace::Filter -- malformed mcm:cpu:thread id '$mct'"
    }

    if {[info exists override($mct)]} {
        return
    }

    if {$dispatch_filter($mct) && $priv_filter($mct)} {
        if {$debug} {
            puts "QTrace::Filter \[$mct\] RESUME qtrace"
        }
        QTrace::Resume $t $c
    } else {
        if {$debug} {
            puts "QTrace::Filter \[$mct\] suspend qtrace"
        }
        QTrace::Pause $t $c
    }
}

#------------------------------------------------------------------------
# QTrace::Dispatch (trigger)
#------------------------------------------------------------------------
proc QTrace::Dispatch { args } {
    variable dispatch_filter
    variable dispatch_focus
    variable debug

    array set p $args
    set sim $p(sim)
    set mct $p(mct)
    set exec $p(exec)

    if {[string match $dispatch_focus($mct) $exec]} {
        set dispatch_filter($mct) 1
    } elseif {$dispatch_filter($mct)} {
        set dispatch_filter($mct) 0
    }

    if {$debug} {
        puts "QTrace::Dispatch mct $mct exec $exec"
    }
    QTrace::Filter $mct
}

#------------------------------------------------------------------------
# QTrace::Focus
#------------------------------------------------------------------------
proc QTrace::Focus { sim mct {f none} } {
    variable dispatch_filter
    variable dispatch_focus
    variable dispatch_callback_id

    switch -exact -- $f {
        none {
            if {[info exists dispatch_callback_id($mct)]} {
                BCast::Ignore $dispatch_callback_id($mct) exec
            }
            set dispatch_filter($mct) 1
            set dispatch_focus($mct) "none"
            puts "QTrace::Focus -- \[$mct\] Remove process filters"
        }
        default {
            if {![info exists dispatch_callback_id($mct)]} {
                set dispatch_callback_id($mct) [BCast::Register $mct [list QTrace::Dispatch sim $sim]]
            }
            BCast::Listen exec $dispatch_callback_id($mct)
            set dispatch_filter($mct) 0
            set dispatch_focus($mct) $f
            puts "QTrace::Focus -- \[$mct\] Filter on process $f"
        }
    }
}

#------------------------------------------------------------------------
# QTrace::Mode_Switch (trigger)
#------------------------------------------------------------------------
proc QTrace::Mode_Switch {args} {
    variable priv_filter
    variable priv_mode
    variable debug

    array set p $args
    set sim $p(sim)
    set msr $p(val)
    set mcm $p(mcm)
    set cpu $p(cpu)
    set thread $p(thread)
    set mct "$mcm:$cpu:$thread"

    if {[info exists priv_mode($mct)]} {
        switch -exact -- $priv_mode($mct) {
            user {
                set priv_filter($mct) [expr ($msr & 0x4000) != 0]
            }
            kernel {
                set priv_filter($mct) [expr ($msr & 0x4000) == 0]
            }
        }

        if {$debug} {
            puts "QTrace::Mode_Switch mct $mct enable $priv_filter($mct)"
        }
        QTrace::Filter $mct
    }
}

#------------------------------------------------------------------------
# QTrace::Priv_Mode
#------------------------------------------------------------------------
proc QTrace::Priv_Mode { sim mct {mode none} } {
    variable dispatch_filter
    variable priv_filter
    variable priv_mode
    variable priv_trig_id

    if {![info exists dispatch_filter($mct)]} {
        set dispatch_filter($mct) 1
    }

    switch -regexp -- $mode {
        none {
            if {[info exists priv_trig_id($mct)]} {
                $sim cpu $mct trigger spr msr id $priv_trig_id($mct) change
                unset priv_trig_id($mct)
            }
            set priv_filter($mct) 1
            set priv_mode($mct) "none"
            puts "QTrace::Priv_Mode -- \[$mct\] Remove priviledge mode filters"
        }
        "user\|kernel" {
            if {![info exists priv_trig_id($mct)]} {
                set priv_trig_id($mct) [$sim cpu $mct trigger spr msr change [list QTrace::Mode_Switch sim $sim] 0x4000]
            }
            set priv_filter($mct) 0
            set priv_mode($mct) $mode
            puts "QTrace::Priv_Mode -- \[$mct\] Filter on priviledge mode $mode"
            QTrace::Sync_Filter $sim $mct
        }
    }
}

#------------------------------------------------------------------------
# QTrace::Force_Filter
#------------------------------------------------------------------------
proc QTrace::Force_Filter {mct enable} {
    variable dispatch_filter
    variable priv_filter

    if {![regexp {((\d):)?(\d):(\d)} $mct match x m c t]} {
        return -code error "QTrace::Force_Filter -- malformed mcm:cpu:thread id '$mct'"
    }

    if {$enable} {
        puts "QTrace::Force_Filter -- forcing ON"
        set dispatch_filter($mct) 1
        set priv_filter($mct) 1
        QTrace::Resume $t $c
    } else {
        puts "QTrace::Force_Filter -- forcing off"
        set dispatch_filter($mct) 0
        set priv_filter($mct) 0
        QTrace::Pause $t $c
    }
}

#------------------------------------------------------------------------
# QTrace::Sync_Filter
#------------------------------------------------------------------------
proc QTrace::Sync_Filter {sim mct} {
    variable dispatch_filter
    variable priv_filter
    variable priv_mode

    if {![regexp {((\d):)?(\d):(\d)} $mct match x m c t]} {
        return -code error "QTrace::Sync_Filter -- malformed mcm:cpu:thread id '$mct'"
    }

    if {[info exists priv_mode($mct)]} {
        set msr    [$sim cpu $c thread $t display spr msr]
        switch -exact -- $priv_mode($mct) {
            user {
                set enable [expr ($msr & 0x4000) != 0]
            }
            kernel {
                set enable [expr ($msr & 0x4000) == 0]
            }
        }
    }

    if {![info exists dispatch_filter($mct)]} {
        set dispatch_filter($mct) 1
    }

    if {$enable} {
        puts "QTrace::Sync_Filter -- forcing ON (in desired mode $priv_mode($mct))"
        set priv_filter($mct) 1
        QTrace::Resume $t $c
    } else {
        puts "QTrace::Sync_Filter -- forcing off (not in desired mode $priv_mode($mct))"
        set priv_filter($mct) 0
        QTrace::Pause $t $c
    }
}

#------------------------------------------------------------------------
# QTrace::Filter_Override
#------------------------------------------------------------------------
proc QTrace::Filter_Override {mct enable} {
    variable override

    if {![regexp {((\d):)?(\d):(\d)} $mct match x m c t]} {
        return -code error "QTrace::Filter_Override -- malformed mcm:cpu:thread id '$mct'"
    }

    if {$enable} {
        puts "QTrace::Filter_Override -- trace resumed, filtering disabled"
        QTrace::Resume $t $c
    } else {
        puts "QTrace::Filter_Override -- trace paused, filtering disabled"
        QTrace::Pause $t $c
    }
    set override($mct) $enable
}

#------------------------------------------------------------------------
# QTrace::Filter_Automatic
#------------------------------------------------------------------------
proc QTrace::Filter_Automatic {mct} {
    variable override
    variable dispatch_filter
    variable priv_filter

    if {![regexp {((\d):)?(\d):(\d)} $mct match x m c t]} {
        return -code error "QTrace::Filter_Automatic -- malformed mcm:cpu:thread id '$mct'"
    }

    if {![info exists dispatch_filter($mct)]} {
        set dispatch_filter($mct) 1
    }

    if {![info exists priv_filter($mct)]} {
        set priv_filter($mct) 1
    }

    unset override($mct)
    QTrace::Filter $mct
}
